/*-------------------------------------------------------------------------
 *
 * pg_explaintips.c
 *	  allow EXPLAIN to dump even more details
 *
 * Copyright (c) 2016-2025, PostgreSQL Global Development Group
 *
 *	  contrib/pg_explaintips/pg_explaintips.c
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include "catalog/pg_class.h"
#include "commands/defrem.h"
#include "commands/explain.h"
#include "commands/explain_format.h"
#include "commands/explain_state.h"
#include "fmgr.h"
#include "parser/parsetree.h"
#include "storage/lock.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"

PG_MODULE_MAGIC;

typedef struct
{
	bool		tips;
} explaintips_options;

static explaintips_options *explaintips_ensure_options(ExplainState *es);
static void explaintips_handler(ExplainState *, DefElem *, ParseState *);
static void explaintips_per_node_hook(PlanState *planstate, List *ancestors,
									  const char *relationship,
									  const char *plan_name,
									  ExplainState *es);

static int	es_extension_id;
static explain_per_node_hook_type prev_explain_per_node_hook;

/*
 * Initialization we do when this module is loaded.
 */
void
_PG_init(void)
{
	/* Get an ID that we can use to cache data in an ExplainState. */
	es_extension_id = GetExplainExtensionId("pg_explaintips");

	/* Register the new EXPLAIN options implemented by this module. */
	RegisterExtensionExplainOption("tips", explaintips_handler);

	/* Use the per-node and per-plan hooks to make our options do something. */
	prev_explain_per_node_hook = explain_per_node_hook;
	explain_per_node_hook = explaintips_per_node_hook;
}

/*
 * Get the explaintips_options structure from an ExplainState; if there is
 * none, create one, attach it to the ExplainState, and return it.
 */
static explaintips_options *
explaintips_ensure_options(ExplainState *es)
{
	explaintips_options *options;

	options = GetExplainExtensionState(es, es_extension_id);

	if (options == NULL)
	{
		options = palloc0(sizeof(explaintips_options));
		SetExplainExtensionState(es, es_extension_id, options);
	}

	return options;
}

/*
 * Parse handler for EXPLAIN (DEBUG).
 */
static void
explaintips_handler(ExplainState *es, DefElem *opt, ParseState *pstate)
{
	explaintips_options *options = explaintips_ensure_options(es);

	options->tips = defGetBoolean(opt);
}

/*
 * Print out additional per-node information as appropriate. If the user didn't
 * specify any of the options we support, do nothing; else, print whatever is
 * relevant to the specified options.
 */
static void
explaintips_per_node_hook(PlanState *planstate, List *ancestors,
						  const char *relationship, const char *plan_name,
						  ExplainState *es)
{
	StringInfoData flags;
	explaintips_options *options;
	Plan	   *plan = planstate->plan;

	options = GetExplainExtensionState(es, es_extension_id);
	if (options == NULL)
		return;

	/*
	 * If the "debug" option was given, display miscellaneous fields from the
	 * "Plan" node that would not otherwise be displayed.
	 */
	if (options->tips)
	{
		/* if seqscan and filteredrows/totalsrows>70) --> tips */
		if (nodeTag(plan) == T_SeqScan)
		{
			double  rows = planstate->instrument->ntuples;
			double  nfiltered = planstate->instrument->nfiltered1;
			if (100*nfiltered/(nfiltered+rows) > 70)
			{
				initStringInfo(&flags);
				appendStringInfo(&flags, "You should probably add an index!");
				ExplainPropertyText("Tips", flags.data, es);
			}
		}
	}
}

